#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: hspylib
   @package: hspylib.modules.fetch
      @file: uri_builder.py
   @created: Mon, 12 Dec 2022
    @author: "<B>H</B>ugo <B>S</B>aporetti <B>J</B>unior")"
      @site: "https://github.com/yorevs/hspylib")
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

from collections import namedtuple
from hspylib.modules.fetch.uri_scheme import UriScheme
from typing import Any, Dict, List
from urllib.parse import parse_qs, SplitResult, urlencode, urlsplit, urlunparse

URI = namedtuple(typename="URI", field_names=["scheme", "netloc", "url", "path", "query", "fragment"])


class UriBuilder:
    @classmethod
    def ensure_scheme(cls, url_string: str, scheme: str | UriScheme = UriScheme.HTTP) -> str:
        if not url_string.startswith(tuple(UriScheme.values())):
            url_string = f"{scheme if isinstance(scheme, str) else str(scheme)}://{url_string}"
        return url_string

    @classmethod
    def of(cls, url_string: str) -> "UriBuilder":
        uri_parts: SplitResult = urlsplit(cls.ensure_scheme(url_string))
        return (
            UriBuilder()
            .scheme(uri_parts.scheme or UriScheme.HTTP)
            .hostname(uri_parts.hostname)
            .port(uri_parts.port)
            .query({k: v[0] for k, v in parse_qs(uri_parts.query).items()})
            .user_info(uri_parts.username, uri_parts.password)
            .add_path(uri_parts.path.split("/"))
            .fragment(uri_parts.fragment)
        )

    def __init__(self) -> None:
        self._scheme: UriScheme = UriScheme.HTTP
        self._host: str = "localhost"
        self._port: int = 80
        self._username: str = ""
        self._password: str = ""
        self._path: List[str] = []
        self._query: Dict[str, Any] = {}
        self._fragment: str = ""

    def __str__(self) -> str:
        return str(self.get_uri())

    def __repr__(self) -> str:
        return str(self)

    def scheme(self, scheme: str | UriScheme) -> "UriBuilder":
        """TODO"""
        self._scheme = scheme if isinstance(scheme, UriScheme) else UriScheme.of(scheme)
        return self

    def hostname(self, host: str) -> "UriBuilder":
        """TODO"""
        self._host = host
        return self

    def port(self, port: int) -> "UriBuilder":
        """TODO"""
        self._port = port
        return self

    def user_info(self, username: str, password: str | None = None) -> "UriBuilder":
        """TODO"""
        if username:
            self._username = username
            self._password = password
        return self

    def add_path(self, path: str | List[str]) -> "UriBuilder":
        """TODO"""
        self._path.extend(path if isinstance(path, List) else [path])
        return self

    def path(self, path: str) -> "UriBuilder":
        """TODO"""
        self._path = [path]
        return self

    def add_query(self, key: str, value: Any) -> "UriBuilder":
        """TODO"""
        self._query.update({key: value})
        return self

    def query(self, query: Dict[str, Any]) -> "UriBuilder":
        """TODO"""
        self._query = query
        return self

    def fragment(self, fragment: str) -> "UriBuilder":
        """TODO"""
        self._fragment = fragment
        return self

    def build(self) -> str:
        """TODO"""
        return urlunparse(self.get_uri())

    def get_uri(self) -> URI:
        return URI(
            scheme=str(self._scheme),
            netloc=self.get_netloc(),
            url="/".join(self._path),
            query=urlencode(self._query),
            path="",
            fragment=self._fragment,
        )

    def get_netloc(self) -> str:
        return (
            f"{self._username}{':' + self._password if self._password else ''}{'@' if self._username else ''}"
            f"{self._host}{':' + str(self._port) if self._port else ''}"
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: hspylib.modules.eventbus
      @file: event.py
   @created: Wed, 30 Jun 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

from hspylib.core.namespace import Namespace


class Event:
    """TODO"""

    def __init__(self, event_name: str, **kwargs):
        self._name = event_name
        self._args = Namespace("EventArgs", True, **kwargs)

    def __str__(self) -> str:
        return f"Event(name={self.name})"

    def __repr__(self):
        return str(self)

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other: "Event") -> bool:
        if isinstance(other, self.__class__):
            return self.name == other.name
        return NotImplemented

    def __getitem__(self, item: str):
        return getattr(self, item)

    @property
    def name(self) -> str:
        return self._name

    @property
    def args(self) -> object:
        return self._args



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: hspylib.modules.application
      @file: exit_status.py
   @created: Thu, 20 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

from hspylib.core.enums.enumeration import Enumeration
from hspylib.modules.cli.vt100.vt_color import VtColor


class ExitStatus(Enumeration):
    """The exit status of an executed command is the value returned by the waitpid system call or equivalent function.
    Exit statuses fall between 0 and 255, though, as explained below, the shell may use values above 125 specially.
    Exit statuses from shell builtins and compound commands are also limited to this range. Under certain circumstances,
    the shell will use special values to indicate specific failure modes."""

    # fmt: off

    # Returned when something ran successfully without errors.
    NOT_SET     = None

    # Returned when something ran successfully without errors.
    SUCCESS     = 0

    # Returned when something went wrong due to any Human interaction.
    ERROR       = 1

    # Returned when something that was supposed to work and failed due to unexpected software behaviour.
    FAILED      = 2

    # Returned when something ran should be aborted. Generally due to user cancellation.
    ABORTED     = 127

    # Returned when something unexpected occurred.
    ABNORMAL    = 129

    # fmt: on

    @staticmethod
    def of(value: int | SystemExit) -> "ExitStatus":
        try:
            return ExitStatus.of_value(value.code if isinstance(value, SystemExit) else value or 0)
        except TypeError:
            return ExitStatus.ABNORMAL

    def __str__(self):
        return self.name

    def __repr__(self):
        color = VtColor.GREEN.code if self == ExitStatus.SUCCESS else VtColor.RED.code
        return f"{color}{self.name}({self.value}){VtColor.NC.code}"

    @property
    def val(self) -> int:
        return int(str(self.value))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: hspylib.core.config
      @file: app_config.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

from hspylib.core.config.properties import Properties
from hspylib.core.metaclass.singleton import Singleton
from hspylib.core.preconditions import check_argument
from hspylib.core.tools.commons import run_dir, str_to_bool
from textwrap import dedent
from typing import Any, Optional

import logging as log
import os


class AppConfigs(metaclass=Singleton):
    """Holds all of the application configurations (properties)"""

    INSTANCE = None

    DISPLAY_FORMAT = dedent(
        """
    AppConfigs
      |-Run-dir = {}
      |-Resources-dir = {}
      |-Properties:
       \\-{}
    """
    )

    def __init__(self, resource_dir: str, filename: str | None = None, profile: str | None = None):
        check_argument(os.path.exists(resource_dir), "Unable to locate resources dir: {}", resource_dir)
        self._resource_dir = resource_dir
        self._properties = Properties(filename=filename, load_dir=resource_dir, profile=profile)
        log.info(self)

    # pylint: disable=consider-using-f-string
    def __str__(self) -> str:
        return "\n{}{}{}".format(
            "-=" * 40,
            self.DISPLAY_FORMAT.format(
                str(run_dir()),
                str(self._resource_dir),
                str(self._properties).replace(os.linesep, f"{os.linesep}   |-") if self._properties.size > 0 else "",
            ),
            "-=" * 40,
        )

    def __repr__(self):
        return str(self)

    def __getitem__(self, item: str) -> Optional[str]:
        return self.get(item)

    def __len__(self) -> int:
        return self.size

    @property
    def resource_dir(self) -> Optional[str]:
        """Return the configured application resource dir"""
        return self._resource_dir

    @property
    def properties(self) -> Properties:
        """Return the application properties"""
        return self._properties

    @property
    def size(self) -> int:
        """Return the application properties"""
        return self._properties.size

    def get(self, property_name: str) -> Optional[Any]:
        """Get the value, as a string, of a property specified by property_name, otherwise None is returned"""
        return self._properties.get(property_name)

    def get_int(self, property_name: str) -> Optional[int]:
        """Get the value, as an integer, of a property specified by property_name, otherwise None is returned"""
        return self._properties.get(property_name, cb_to_type=int)

    def get_float(self, property_name: str) -> Optional[float]:
        """Get the value, as a float, of a property specified by property_name, otherwise None is returned"""
        return self._properties.get(property_name, cb_to_type=float)

    def get_bool(self, property_name: str) -> Optional[bool]:
        """Get the value, as a boolean, of a property specified by property_name, otherwise None is returned"""
        return self._properties.get(property_name, cb_to_type=str_to_bool)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: hspylib.modules.fetch
      @file: fetch.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
from hspylib.core.enums.http_method import HttpMethod
from hspylib.core.tools.commons import sysout
from hspylib.modules.fetch.http_response import HttpResponse
from hspylib.modules.fetch.uri_builder import UriBuilder
from retry import retry
from typing import Any, Dict, List, Tuple, Union
from urllib3.exceptions import NewConnectionError
from urllib.error import HTTPError

import logging as log
import requests
import requests.exceptions as exs

RETRYABLE_EXS = (
    NewConnectionError,
    HTTPError,
    exs.ConnectTimeout,
    exs.ConnectionError,
    exs.ReadTimeout,
    exs.InvalidURL,
    exs.InvalidSchema,
)


@retry(exceptions=RETRYABLE_EXS, tries=3, delay=1, backoff=3, max_delay=30, jitter=0.75)
def fetch(
    url: str,
    method: HttpMethod = HttpMethod.GET,
    headers: List[Dict[str, str]] = None,
    body: Any = None,
    silent: bool = True,
    timeout: Union[float, Tuple[float, float]] = 10,
) -> HttpResponse:
    """Do a request specified by method and according to parameters.
    :param url: The url to make the request.
    :param method: The http method to be used [ GET, HEAD, POST, PUT, PATCH, DELETE, OPTIONS ].
    :param headers: The http request headers.
    :param body: The http request body (payload).
    :param silent: Omits all informational messages.
    :param timeout: How many seconds to wait for the server to send data or connect before giving up.
    :return:
    """

    final_url = UriBuilder.ensure_scheme(url)
    if not silent:
        sysout(
            f"Fetching: "
            f"method={method} headers={headers if headers else '[]'} "
            f"body={body if body else '{}'} url={final_url} ..."
        )

    all_headers = {}
    if headers:
        list(map(all_headers.update, headers))

    response = requests.request(
        url=final_url, method=method.name, headers=all_headers, data=body, timeout=timeout, verify=False
    )

    return HttpResponse.of(response)


def head(
    url: str, headers: List[Dict[str, str]] = None, silent: bool = True, timeout: Union[float, Tuple[float, float]] = 10
) -> HttpResponse:
    """Do HEAD request and according to parameters."""

    return fetch(url=url, method=HttpMethod.HEAD, headers=headers, silent=silent, timeout=timeout)


def get(
    url: str, headers: List[Dict[str, str]] = None, silent: bool = True, timeout: Union[float, Tuple[float, float]] = 10
) -> HttpResponse:
    """Do GET request and according to parameters."""

    return fetch(url=url, headers=headers, silent=silent, timeout=timeout)


def delete(
    url: str, headers: List[Dict[str, str]] = None, silent: bool = True, timeout: Union[float, Tuple[float, float]] = 10
) -> HttpResponse:
    """Do DELETE request and according to parameters."""

    return fetch(url=url, method=HttpMethod.DELETE, headers=headers, silent=silent, timeout=timeout)


def post(
    url: str,
    body=None,
    headers: List[Dict[str, str]] = None,
    silent: bool = True,
    timeout: Union[float, Tuple[float, float]] = 10,
) -> HttpResponse:
    """Do POST request and according to parameters."""

    return fetch(url=url, method=HttpMethod.POST, headers=headers, body=body, silent=silent, timeout=timeout)


def put(
    url: str,
    body=None,
    headers: List[Dict[str, str]] = None,
    silent: bool = True,
    timeout: Union[float, Tuple[float, float]] = 10,
) -> HttpResponse:
    """Do PUT request and according to parameters."""

    return fetch(url=url, method=HttpMethod.PUT, headers=headers, body=body, silent=silent, timeout=timeout)


def patch(
    url: str,
    body=None,
    headers: List[Dict[str, str]] = None,
    silent: bool = True,
    timeout: Union[float, Tuple[float, float]] = 10,
) -> HttpResponse:
    """Do PATCH request and according to parameters."""

    return fetch(url=url, method=HttpMethod.PATCH, headers=headers, body=body, silent=silent, timeout=timeout)


def is_reachable(urls: str | Tuple[str], timeout: Union[float, Tuple[float, float]] = 1) -> bool:
    """Check if the specified url is reachable"""
    reachable = True

    try:
        if isinstance(urls, Tuple):
            return all(is_reachable(UriBuilder.ensure_scheme(u)) for u in urls)
        requests.options(url=UriBuilder.ensure_scheme(urls), timeout=timeout)
    except RETRYABLE_EXS as err:
        log.warning("URLs %s is not reachable => %s", urls, err)
        reachable = False

    return reachable



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: hspylib.modules.application.argparse
      @file: argument_parser.py
   @created: Thu, 14 Feb 2022
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
from argparse import ArgumentParser
from hspylib.modules.application.exit_status import ExitStatus
from typing import Any

import argparse
import sys


class HSArgumentParser(ArgumentParser):
    """HSPyLib custom argument parser"""

    def _check_value(self, action, value: Any):
        if action.choices is not None and value not in action.choices:
            msg = f"Invalid choice '{value}'. Choose from [{', '.join(map(repr, action.choices))}]"
            raise argparse.ArgumentError(action, msg)

    def error(self, message):
        self.print_help(sys.stderr)
        self.exit(ExitStatus.FAILED.val, f"\n### Error {self.prog} -> {message}\n\n")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: hspylib.core
      @file: namespace.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
from hspylib.core.preconditions import check_not_none
from hspylib.core.tools.dict_tools import merge
from typing import Any, Dict, Iterator, List, Optional, Tuple

ATTRIBUTE_TYPES = Dict[str, Any] | Tuple[Dict[str, Any]] | List[Dict[str, Any]]


class Namespace:
    """TODO"""

    @staticmethod
    def of(type_name: str, attributes: ATTRIBUTE_TYPES, final: bool = False) -> "Namespace":
        check_not_none(attributes)
        self = Namespace(type_name, final)
        self += attributes if isinstance(attributes, dict) else merge(attributes)
        return self

    def __init__(self, type_name: str = None, final: bool = False, **kwargs) -> None:
        self.__name__ = type_name or self.__class__.__name__
        self._final = None
        self._index = 0
        list(map(self.setattr, kwargs.keys(), kwargs.values()))
        self._final = final

    @property
    def attributes(self) -> Tuple[str]:
        # fmt: off
        return tuple(
            attr for attr in filter(
                lambda name:
                    self[name] is not None
                    and not name.startswith(("_", "__"))
                    and not name.endswith(("_", "__")),
                vars(self),
            )
        )
        # fmt: on

    @property
    def values(self) -> Tuple[Any]:
        all_values = [self[attr] for attr in self.attributes]
        # fmt: off
        return tuple(
            filter(
                lambda val:
                    val is not None,
                all_values
            )
        )
        # fmt: on

    def __str__(self) -> str:
        return f"{self.__name__}({', '.join([f'{a}={av}' for a, av in zip(self.attributes, self.values)])})"

    def __repr__(self) -> str:
        return str(self)

    def __hash__(self) -> int:
        return hash(self.key())

    def __eq__(self, other: "Namespace") -> bool:
        if isinstance(other, self.__class__):
            return self.key() == other.key()
        return NotImplemented

    def __getitem__(self, attribute_name: str) -> Any:
        return getattr(self, attribute_name)

    def __len__(self) -> int:
        return len(self.attributes)

    def __iter__(self) -> Iterator[Tuple[str, Any]]:
        self._index = 0
        return self

    def __next__(self) -> Tuple[str, Any]:
        if self._index < len(self):
            item = self.item_at(self._index)
            self._index += 1
            return item
        raise StopIteration

    def __contains__(self, *attributes: str):
        return all(hasattr(self, attr) for attr in attributes)

    def __iadd__(self, attribute: Dict[str, Any] | "Namespace") -> "Namespace":
        for a, av in attribute.items():
            self.setattr(a, av)
        return self

    def __add__(self, other: Dict[str, Any] | "Namespace") -> "Namespace":
        self.__name__ += "_" + other.__name__
        return self.__iadd__(other)

    def key(self) -> Tuple[Any]:
        return tuple(self.values)

    def setattr(self, name: str, value: Any) -> "Namespace":
        """TODO"""
        if self._final and not hasattr(self, name):
            raise ValueError(f"Can't set attribute '{name}'. '{self.__name__}' Namespace is final")
        if not hasattr(self, name):
            self._check_name(name)
        setattr(self, name, value)
        return self

    def item_at(self, index: int) -> Optional[Tuple[str, Any]]:
        """TODO"""
        # fmt: off
        return (
            self.attributes[index], self.values[index]
        ) if index < len(self) \
          else None
        # fmt: on

    def items(self) -> Iterator[Tuple[str, Any]]:
        """TODO"""
        return iter(self)

    def hasattr(self, *names: str) -> bool:
        """TODO"""
        return self.__contains__(*names)

    def _asdict(self) -> Dict[str, Any]:
        """
        To match the same function name as namedtuple, we used _asdict
        """
        return dict(zip(self.attributes, self.values))

    def _check_name(self, name: str) -> None:
        """TODO"""
        # fmt: off
        if not (
            self.hasattr(name)
            or (
                not name.startswith(("_", "__"))
                and not name.endswith(("_", "__"))
                and name not in (forbidden := vars(self).keys())
            )
        ):
            raise NameError(
                f"Invalid attribute name '{name}'. "
                f"Attributes can't start with '_' or '__' and must not be in {forbidden}")
        # fmt: on



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: hspylib.modules.qt.promotions
      @file: hlistwidget.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
from hspylib.core.preconditions import check_argument, check_not_none, check_state
from PyQt5.QtCore import pyqtSignal, QModelIndex, Qt
from PyQt5.QtGui import QCursor, QKeyEvent
from PyQt5.QtWidgets import QListWidget, QListWidgetItem, QMenu, QWidget
from typing import List, Optional, Union


class HListWidget(QListWidget):
    """TODO"""

    keyPressed = pyqtSignal(int)

    @staticmethod
    def unset_flag(flags: int, flag: int) -> int:
        """TODO"""
        return flags ^ flag if flags & flag == flag else flags

    def __init__(self, parent: Optional[QWidget] = None):
        super().__init__(parent)
        self._editable = False
        self._selectable = True
        self._items = []
        self._custom_menu_actions = []
        self._context_menu_enable = True
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self._context_menu)
        self.itemChanged.connect(self.item_changed)

    def keyPressEvent(self, event: QKeyEvent) -> None:
        """Handles and forwards the key press event in the list."""
        super().keyPressEvent(event)
        self.keyPressed.emit(event.key())

    def item_changed(self, item: QListWidgetItem):
        """Handles itemChanged event to avoid adding duplicates to the list."""
        existing = self.findItems(item.text(), Qt.MatchFixedString)
        if len(existing) > 1:
            self.del_item(item)

    def addItem(self, item: QListWidgetItem) -> None:
        """Adds the specified element to this list regardless if it is present or not."""
        prev = self.item(self.count() - 1)
        if prev:
            item.setFlags(prev.flags())
        super().addItem(item)
        self._items.append(item)

    def is_empty(self) -> bool:
        """Returns true if this list contains no elements."""
        return len(self._items) == 0

    def set_item(self, item: Union[str, QListWidgetItem], flags: Union[Qt.ItemFlags, Qt.ItemFlag] = None) -> None:
        """Adds the specified element to this list if it is not already present."""
        if not self.findItems(item, Qt.MatchFixedString):
            if isinstance(item, str):
                w_item = QListWidgetItem(item)
                w_item.setFlags(flags or Qt.ItemIsSelectable | Qt.ItemIsEditable | Qt.ItemIsEnabled)
                self.addItem(w_item)
            else:
                item.setFlags(flags or Qt.ItemIsSelectable | Qt.ItemIsEditable | Qt.ItemIsEnabled)
                self.addItem(item)

    def remove_item_at_index(self, row: int) -> Optional[QListWidgetItem]:
        """Removes the element at the specified row from this list if it is present."""
        item = self.takeItem(row)
        if item:
            self._items.remove(self._items[row])
        return item

    def del_item(self, item_or_index: Union[int, QModelIndex, QListWidgetItem]) -> Optional[QListWidgetItem]:
        """Removes the specified element from this list if it is present."""
        item = None
        if isinstance(item_or_index, int):
            if 0 <= item_or_index < len(self._items):
                item = self.remove_item_at_index(item_or_index)
        elif isinstance(item_or_index, QListWidgetItem):
            index = self.indexFromItem(item_or_index)
            if index:
                item = self.remove_item_at_index(index.row())
        elif isinstance(item_or_index, QModelIndex):
            item = self.remove_item_at_index(item_or_index.row())

        return item

    def clear(self) -> None:
        """Removes all of the elements from this list."""
        super().clear()
        del self._items[:]

    def size(self) -> int:
        """Returns the number of elements in this list."""
        return self.count()

    def index_of(self, item: str) -> int:
        """Returns the index of the first occurrence of the specified element in this list,
        or -1 if this list does not contain the element."""
        return next((obj for obj in self._items if obj == item), -1)

    def as_list(self) -> List[str]:
        """Returns a list containing all items in this list as a string."""
        return list(map(QListWidgetItem.text, self._items))

    def current_text(self) -> str:
        """Returns the current item's text."""
        return self.currentItem().text()

    def set_context_menu_enable(self, enabled: bool = True) -> None:
        """Whether context menu is enabled or not"""
        self._context_menu_enable = enabled

    def set_editable(self, editable: bool = True) -> None:
        """Set editable property."""
        for item in self._items:
            flags = int(item.flags())
            if editable:
                item.setFlags(flags | Qt.ItemIsEditable)
            else:
                item.setFlags(self.unset_flag(flags, int(Qt.ItemIsEditable)))
        self._editable = editable

    def set_selectable(self, selectable: bool = True) -> None:
        """Set selectable property."""
        for item in self._items:
            flags = int(item.flags())
            if selectable:
                item.setFlags(flags | Qt.ItemIsSelectable)
            else:
                item.setFlags(self.unset_flag(flags, int(Qt.ItemIsSelectable)))
        self._selectable = selectable

    def _context_menu(self) -> None:
        """Display the custom context menu."""
        if self._context_menu_enable:
            ctx_menu = QMenu(self)
            if self._editable:
                ctx_menu.addAction("Add Item", lambda: self.set_item("<new_item>"))
                if not self.is_empty():
                    ctx_menu.addAction("Delete Item", lambda: self.del_item(self.currentIndex().row()))
                    ctx_menu.addSeparator()
                    ctx_menu.addAction("Clear list", self.clear)

            for act in self._custom_menu_actions:
                check_not_none(act)
                check_state(len(act) == 3, f"Invalid custom menu action: {act}")
                check_argument(callable(act[1]), "The action must be callable")
                if act[2]:
                    ctx_menu.addSeparator()
                ctx_menu.addAction(act[0], act[1])

            ctx_menu.exec_(QCursor.pos())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: hspylib
   @package: hspylib.core.metaclass
      @file: __classpath__.py
   @created: Wed, 8 Jun 2022
    @author: "<B>H</B>ugo <B>S</B>aporetti <B>J</B>unior")"
      @site: "https://github.com/yorevs/hspylib")
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""
from hspylib.core.enums.charset import Charset
from hspylib.core.exception.exceptions import ResourceNotFoundError, SourceNotFoundError
from hspylib.core.metaclass.singleton import Singleton
from hspylib.core.preconditions import check_argument, check_not_none, check_state
from pathlib import Path
from textwrap import dedent
from typing import Optional

import logging as log
import os

AnyPath = Path | str | None


class Classpath(metaclass=Singleton):
    """The classpath tells Python applications where to look in the filesystem for source and resource files."""

    INSTANCE = None

    def __init__(self, source_root: AnyPath = None, run_dir: AnyPath = None, resource_dir: AnyPath = None):

        if source_root:
            check_state(Path(str(source_root)).exists(), "source_root must exist")
        if run_dir:
            check_state(Path(str(run_dir)).exists(), "run_dir must exist")

        self.source_root = Path(os.getenv("SOURCE_ROOT", str(source_root or os.curdir)))
        self.run_dir = Path(str(run_dir or self.source_root))
        self.resource_dir = Path(str(resource_dir or f"{self.source_root}/resources"))
        self.log_dir = Path(os.getenv("LOG_DIR", f"{self.run_dir}/log") or self.run_dir)

    @classmethod
    def source_path(cls) -> Path:
        """TODO"""
        return cls.INSTANCE.source_root

    @classmethod
    def run_path(cls) -> Path:
        """TODO"""
        return cls.INSTANCE.run_dir

    @classmethod
    def resource_path(cls) -> Path:
        """TODO"""
        return cls.INSTANCE.resource_dir

    @classmethod
    def log_path(cls) -> Path:
        """TODO"""
        return cls.INSTANCE.log_dir

    @classmethod
    def list_resources(cls) -> str:
        def list_resources_closure(directory: str | Path, depth: int = 4) -> Optional[str]:
            """Walk through resources directory and build a list with all files"""
            if os.path.exists(directory):
                res_str = Classpath._list_files(directory, depth)
                for root, dirs, _ in os.walk(directory):
                    for dirname in dirs:
                        res_str += " " * depth + "|-" + dirname + os.linesep
                        res_str += list_resources_closure(os.path.join(root, dirname), depth + 2) or ""
                return res_str
            return None

        return list_resources_closure(cls.resource_path())

    @staticmethod
    def _list_files(directory: str | Path, depth: int = 4) -> str:
        """Walk through directory and build a list with all files"""
        res_str = ""
        if os.path.exists(directory):
            for filename in os.listdir(directory):
                f = os.path.join(directory, filename)
                if os.path.isfile(f):
                    res_str += " " * depth + "|-" + str(filename) + os.linesep
        return res_str

    @classmethod
    def get_resource_path(cls, resource: str | Path) -> Path:
        """Return the path of the given resource"""
        check_not_none(resource, "Must provide a valid resource path")
        resource = Path(f"{cls.resource_path()}/{str(resource)}")
        if not resource.exists():
            raise ResourceNotFoundError(f"Resource {str(resource)} was not found at {cls.INSTANCE.source_path()}!")
        return resource

    @classmethod
    def get_source_path(cls, source: str | Path) -> Path:
        """Return the path of the given source"""
        check_not_none(source, "Must provide a valid source path")
        filepath = Path(f"{cls.source_path()}/{str(source)}")
        if not filepath.exists():
            raise SourceNotFoundError(f"Source {str(source)} was not found at {cls.INSTANCE.source_path()}!")
        return filepath

    @classmethod
    def load_envs(
        cls,
        prefix: str | None = None,
        suffix: str | None = None,
        load_dir: str | None = None,
        raise_error: bool = False,
    ) -> None:
        """Load environment variables from environment file."""

        load_path = Path(load_dir or f"{cls.source_path()}/env")
        check_argument(load_path.exists())
        files = [f"{load_path}/{prefix or ''}{f'-{suffix}' if suffix else '.env'}", f"{load_path}/.envrc"]
        env_file = next((f for f in files if os.path.exists(f)), None)
        if env_file:
            log.debug("ENVIRONMENT::Loading environment file '%s'", env_file)
            with open(env_file, "r", encoding=Charset.UTF_8.val) as f_env:
                lines = f_env.readlines()
                lines = list(filter(lambda l: l.startswith("export "), filter(None, lines)))
                variables = list(map(lambda x: x.split("=", 1), map(lambda l: l[7:].strip(), lines)))
                for v in variables:
                    log.debug("ENVIRONMENT::Setting environment variable\t'%s'", v[0])
                    os.environ[v[0]] = v[1]
        else:
            if raise_error:
                raise FileNotFoundError(f"None of the environment files: {files} were found!")
            log.warning("ENVIRONMENT::None of the environment files: %s were found", files)

    def __str__(self) -> str:
        return (
            dedent(
                f"""
        Classpath(
          |-cur-working-dir: {os.getcwd()}
          |-source-root: {self.source_path()}
          |-run-dir: {self.run_dir}
          |-log-dir: {self.log_path()}
          |-resource-dir: {self.resource_path()}
        """
            )
            + f"{self.list_resources() or ''})"
        )

    def __repr__(self):
        return str(self)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: hspylib.modules.qt.promotions
      @file: htablemodel.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

from hspylib.core.collection_filter import CollectionFilter, FilterCondition
from hspylib.core.tools.commons import class_attribute_names, class_attribute_values
from PyQt5.QtCore import QAbstractTableModel, QModelIndex, Qt, QVariant
from PyQt5.QtGui import QPalette
from PyQt5.QtWidgets import QTableView
from typing import List, Tuple, Type, TypeVar, Union

import collections

T = TypeVar("T")


class HTableModel(QAbstractTableModel):
    """TODO"""

    def __init__(
        self,
        parent: QTableView,
        clazz: Type,
        headers: List[str] = None,
        table_data: List[T] = None,
        cell_alignments: List[Qt.AlignmentFlag] = None,
        max_rows: int = 500,
    ):

        QAbstractTableModel.__init__(self, parent)
        self._parent = parent
        self._clazz = clazz
        self._data = collections.deque(maxlen=max_rows)
        self._filters = CollectionFilter()
        self._headers = headers or self._headers_by_entity()
        self._cell_alignments = cell_alignments
        self._parent.setModel(self)
        self.push_data(table_data or [])

    def removeRow(self, row: int, parent: QModelIndex = ...) -> bool:  # pylint: disable=unused-argument
        """TODO"""
        if 0 <= row < len(self._data):
            self.beginRemoveRows(QModelIndex(), row, row + 1)
            del self._data[row]
            self.layoutChanged.emit()
            self.endRemoveRows()
            return True
        return False

    def data(self, index: QModelIndex, role: int = ...) -> QVariant:
        """TODO"""
        if role == Qt.DisplayRole:
            row_dict = self._data[index.row()].__dict__
            value = class_attribute_values(row_dict)[index.column()]
            ret_val = QVariant(str(value if value is not None else ""))
        elif role == Qt.TextAlignmentRole:
            ret_val = (
                QVariant(self._cell_alignments[index.column()])
                if self._cell_alignments
                else QVariant(Qt.AlignLeft | Qt.AlignVCenter)
            )
        elif role == Qt.BackgroundColorRole:
            light_gray = self._parent.palette().color(QPalette.Window).lighter(100)
            ret_val = QVariant(light_gray if index.row() % 2 != 0 else "")
        else:
            ret_val = None

        return ret_val or QVariant()

    def headerData(self, section: int, orientation: Qt.Orientation, role: int = ...) -> QVariant:
        """TODO"""
        ret_val = None
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            ret_val = QVariant(self._headers[section].upper()) if len(self._headers) >= section else QVariant("-")
        elif orientation == Qt.Vertical and role == Qt.DisplayRole:
            ret_val = QVariant(str(section))

        return ret_val if ret_val else QVariant()

    def rowCount(self, parent: QModelIndex = ...) -> int:  # pylint: disable=unused-argument
        """TODO"""
        return len(self._data) if self._data else 0

    def columnCount(self, parent: QModelIndex = ...) -> int:  # pylint: disable=unused-argument
        """TODO"""
        return len(self._headers) if self._data else 0

    def sort(self, column: int, order: Qt.SortOrder = ...) -> None:
        """TODO"""
        keys = class_attribute_names(self._clazz)
        self._data = sorted(self._data, key=lambda x: getattr(x, keys[column]), reverse=bool(order))
        self.layoutChanged.emit()

    def append(self, data: T):
        """TODO"""
        if data and not self._filters.should_filter(data):
            self.beginInsertRows(QModelIndex(), self.rowCount(), self.rowCount() + 1)
            self._data.append(data)
            self.endInsertRows()
            self.layoutChanged.emit()

    def apply_filter(
        self, name: str, el_name: str, condition: FilterCondition, el_value: Union[int, str, bool, float]
    ) -> None:
        """TODO"""
        self._filters.apply_filter(name, el_name, condition, el_value)

    def filter(self) -> None:
        """TODO"""
        self._data = self._filters.filter(list(self._data))

    def row(self, index: QModelIndex) -> T:
        """TODO"""
        return self._data[index.row()]

    def column(self, index: QModelIndex) -> T:
        """TODO"""
        row = self.row(index)
        col_name = str(list(vars(row))[index.column()])
        return getattr(row, col_name)

    def push_data(self, data: Union[List[T], T]) -> None:
        """TODO"""
        if data:
            if isinstance(data, list):
                list(map(self.append, data))
            else:
                self.append(data)
            self.layoutChanged.emit()

    def clear(self):
        """TODO"""
        self._data.clear()
        self.layoutChanged.emit()

    def is_empty(self) -> bool:
        """TODO"""
        return len(self._data) == 0

    def refresh_data(self) -> None:
        """TODO"""
        self._data = self._filters.filter(list(self._data))
        self.layoutChanged.emit()

    def remove_rows(self, rows: List[QModelIndex]):
        """TODO"""
        # Because we are using deque, we need to sort DESC to avoid deleting wrong indexes
        for row in sorted(rows, reverse=True):
            self.removeRow(row.row())

    def _headers_by_entity(self) -> List[str]:
        """TODO"""
        attributes = class_attribute_names(self._clazz)
        return [str(x).capitalize() for x in attributes]

    def selected_data(self) -> Tuple[List[QModelIndex], List[T]]:
        """TODO"""
        sel_model = self._parent.selectionModel()
        if sel_model:
            sel_indexes = sel_model.selectedIndexes()
            sel_rows = {[idx.row() for idx in sel_indexes]}
            return sel_indexes, [self._data[row] for row in sel_rows]
        return [], []

    def filters(self) -> CollectionFilter:
        """TODO"""
        return self._filters



# _*_ coding: utf-8 _*_
#
# hspylib v0.11.155
#
# Package: main.hspylib.modules.fetch
"""Package initialization."""

__all__ = [
  'fetch', 
  'http_response', 
  'uri_builder', 
  'uri_scheme'
]
__version__ = '0.11.155'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: hspylib.core.config
      @file: properties.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

from collections import defaultdict
from configparser import ConfigParser
from hspylib.core.enums.charset import Charset
from hspylib.core.tools.commons import run_dir, touch_file
from hspylib.core.tools.dict_tools import flatten_dict
from typing import Any, Callable, Dict, Iterator, List, Optional, TextIO, Type

import logging as log
import os
import re
import toml
import yaml

CONVERSION_FN = Type | Callable[[Any], Any]


class Properties:
    """The Properties class represents a persistent set of properties. Each key and its corresponding value in the
    property list is a string."""

    _default_name: str = "application"
    _default_ext: str = ".properties"

    @staticmethod
    def environ_name(property_name: str) -> str:
        """Retrieve the environment name of the specified property name
        :param property_name: the name of the property using space, dot or dash notations
        """
        return re.sub("[ -.]", "_", property_name).upper()

    @staticmethod
    def read_properties(file_handler: TextIO) -> Dict[str, Any]:
        """Reads properties from properties file (key and element pairs) from the input list."""
        all_lines = list(map(str.strip, filter(None, file_handler.readlines())))
        # fmt: off
        return {
            p[0].strip(): p[1].strip()
            for p in [
                p.split("=", 1) for p in list(
                    filter(lambda l: re.match(r"[a-zA-Z]([.\\-]|\w)* *= *.+", l), all_lines)
                )
            ]
        }
        # fmt: off

    @staticmethod
    def read_cfg_or_ini(file_handler: TextIO) -> Dict[str, Any]:
        """Reads properties from a cfg or ini file (key and element pairs) from the input list."""
        all_lines = list(map(str.strip, filter(None, file_handler.readlines())))
        string = os.linesep.join(all_lines)
        all_cfgs = {}
        cfg = ConfigParser()
        cfg.read_string(string)
        for section in cfg.sections():
            all_cfgs.update(dict(cfg.items(section)))
        return all_cfgs

    @staticmethod
    def read_yaml(file_handler: TextIO) -> Dict[str, Any]:
        """Reads properties from a yaml file (key and element pairs) from the input list."""
        return flatten_dict(yaml.safe_load(file_handler))

    @staticmethod
    def read_toml(file_handler: TextIO) -> Dict[str, Any]:
        """Reads properties from a toml file (key and element pairs) from the input list."""
        return flatten_dict(toml.load(file_handler))

    def __init__(self, filename: str = None, profile: str | None = None, load_dir: str | None = None) -> None:

        self._filename, self._extension = os.path.splitext(
            filename if filename else f"{self._default_name}{self._default_ext}"
        )
        self._profile = profile if profile else os.environ.get("ACTIVE_PROFILE", "")
        self._properties = defaultdict()
        self._load(load_dir or f"{run_dir()}/resources")

    def __str__(self) -> str:
        str_val = ""
        for key, value in self._properties.items():
            str_val += "{}{}={}".format("\n" if str_val else "", key, value)
        return str_val

    def __repr__(self) -> str:
        return str(self)

    def __getitem__(self, item: str) -> Optional[Any]:
        return self.get(item)

    def __iter__(self) -> Iterator:
        return self._properties.__iter__()

    def __len__(self) -> int:
        """Retrieve the amount of properties"""
        return len(self._properties)

    def get(self, prop_name: str, cb_to_type: CONVERSION_FN = str, default: Any | None = None) -> Optional[Any]:
        """Retrieve a property specified by property and cast to the proper type. If the property is not found,
        return the default value."""

        try:
            value = self._get(prop_name)
            return cb_to_type(value) if value else None
        except TypeError:
            log.warning("Unable to convert property '%s' into '%s'", prop_name, cb_to_type)
            return default

    @property
    def values(self) -> List[Any]:
        """Retrieve all values for all properties"""
        return list(self._properties.values())

    @property
    def keys(self) -> List[str]:
        """Retrieve all values for all properties"""
        return list(self._properties.keys())

    @property
    def size(self) -> int:
        """Retrieve the amount of properties actually store."""
        return len(self._properties)

    def _get(self, key: str, default: Any = None) -> Optional[Any]:
        """Get a property value as string or default_value if the property was not found"""
        if value := os.environ.get(self.environ_name(key), None):
            return value
        return self._properties[key] if key in self._properties else default

    def _load(self, load_dir: str) -> None:
        """Read all properties from the file"""
        filepath = self._build_path(load_dir)
        if os.path.exists(filepath):
            return self._parse(filepath)

        raise FileNotFoundError(f'File "{filepath}" does not exist')

    def _build_path(self, load_dir: str) -> str:
        """Find the proper path for the properties file"""
        return f"{load_dir}/{self._filename}{'-' + self._profile if self._profile else ''}{self._extension}"

    def _parse(self, filepath: str) -> None:
        """Parse the properties file according to it's extension"""
        if not os.path.isfile(filepath):
            touch_file(filepath)
        ext = self._extension.lower()
        with open(filepath, encoding=Charset.UTF_8.val) as fh_props:
            if ext in [".ini", ".cfg"]:
                all_properties = self.read_cfg_or_ini(fh_props)
            elif ext == ".properties":
                all_properties = self.read_properties(fh_props)
            elif ext in [".yml", ".yaml"]:
                all_properties = self.read_yaml(fh_props)
            elif ext == ".toml":
                all_properties = self.read_toml(fh_props)
            else:
                raise NotImplementedError(f"Extension {ext} is not supported")
            self._properties.update(all_properties)
        log.debug("Successfully loaded %d properties from: %s", len(self._properties), filepath)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HSPyLib
   @package: hspylib.modules.web
      @file: rest_utils.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

from hspylib.core.enums.http_method import HttpMethod
from hspylib.modules.fetch.fetch import fetch
from hspylib.modules.fetch.http_response import HttpResponse
from requests.structures import CaseInsensitiveDict
from typing import Optional

import logging as log


def rest_call(
    url: str, method: HttpMethod, data: str, headers: Optional[CaseInsensitiveDict]
) -> Optional[HttpResponse]:
    """TODO"""

    all_headers = {} if not headers else headers
    all_headers.update({"Content-Type": "application/json", "Accept": "*/*"})

    log.debug("Processing REST {all_headers} {method} -> {url}")
    response = fetch(url=url, method=method, headers=all_headers, body=data)
    log.debug("Response <=  Status: %s  Payload: %s", response.status_code, response)

    return response



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: hspylib
   @package: hspylib.modules.qt.promotions
      @file: htoolbox.py
   @created: Fri, 29 Jul 2022
    @author: "<B>H</B>ugo <B>S</B>aporetti <B>J</B>unior")"
      @site: "https://github.com/yorevs/hspylib")
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2022, HSPyLib team
"""

from functools import cached_property
from PyQt5 import QtWidgets
from PyQt5.QtCore import pyqtProperty, pyqtSignal, pyqtSlot, QObject, Qt, QVariantAnimation
from PyQt5.QtWidgets import QToolBox


class HToolBox(QToolBox):
    """Animated QToolBox
    Inspired by: https://stackoverflow.com/questions/70746992/animation-effect-in-a-qtoolbox-in-pyqt5-and-python"""

    class _ToolBoxPage(QObject):
        """Animated QToolBox page
        Inspired by: https://stackoverflow.com/questions/70746992/animation-effect-in-a-qtoolbox-in-pyqt5-and-python"""

        QT_MAX_HEIGHT = 16777215

        destroyed = pyqtSignal()

        def __init__(self, button, scroll_area):
            super().__init__()
            self.button = button
            self.scrollArea = scroll_area
            self.widget = scroll_area.widget()
            self.widget.destroyed.connect(self.destroyed)

        def begin_hide(self, spacing):
            """TODO"""
            self.scrollArea.setMinimumHeight(1)
            self.scrollArea.setMaximumHeight(self.scrollArea.height() - spacing)
            if not self.scrollArea.verticalScrollBar().isVisible():
                self.scrollArea.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

        def begin_show(self, target_height):
            """TODO"""
            if self.scrollArea.widget().minimumSizeHint().height() <= target_height:
                self.scrollArea.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
            else:
                self.scrollArea.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
            self.scrollArea.setMaximumHeight(0)
            self.scrollArea.show()

        def set_height(self, height):
            """TODO"""
            if height and not self.scrollArea.minimumHeight():
                self.scrollArea.setMinimumHeight(1)
            self.scrollArea.setMaximumHeight(height)

        def finalize(self):
            """TODO"""
            self.scrollArea.setMinimumHeight(0)
            self.scrollArea.setMaximumHeight(self.QT_MAX_HEIGHT)
            self.scrollArea.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)

    _oldPage = _newPage = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._pages = []
        self._targetSize = None

    @cached_property
    def animation(self):
        """TODO"""
        animation = QVariantAnimation(self)
        animation.setDuration(120)
        animation.setStartValue(0.0)
        animation.setEndValue(1.0)
        animation.valueChanged.connect(self._update_sizes)
        return animation

    @pyqtProperty(int)
    def animationDuration(self):
        return self.animation.duration()

    @animationDuration.setter
    def animationDuration(self, duration):
        self.animation.setDuration(max(50, min(duration, 500)))

    @pyqtSlot(int)
    @pyqtSlot(int, bool)
    def setCurrentIndex(self, index, now=False):
        if self.currentIndex() == index:
            return
        if now:
            if self.animation.state():
                self.animation.stop()
                self._pages[index].finalize()
            super().setCurrentIndex(index)
            return
        if self.animation.state():
            return

        self._oldPage = self._pages[self.currentIndex()]
        self._oldPage.begin_hide(self.layout().spacing())
        self._newPage = self._pages[index]
        self._newPage.begin_show(self._targetSize)
        self.animation.start()

    @pyqtSlot(QtWidgets.QWidget)
    @pyqtSlot(QtWidgets.QWidget, bool)
    def setCurrentWidget(self, widget):
        for i, page in enumerate(self._pages):
            if page.widget == widget:
                self.setCurrentIndex(i)
                return

    def itemInserted(self, index):
        button = self.layout().itemAt(index * 2).widget()
        button.clicked.disconnect()
        button.clicked.connect(self._button_clicked)
        scroll_area = self.layout().itemAt(index * 2 + 1).widget()
        page = HToolBox._ToolBoxPage(button, scroll_area)
        self._pages.insert(index, page)
        page.destroyed.connect(self._widget_destroyed)
        self._compute_target_size()

    def itemRemoved(self, index):
        if self.animation.state() and self._index(self._newPage) == index:
            self.animation.stop()
        page = self._pages.pop(index)
        page.destroyed.disconnect(self._widget_destroyed)
        self._compute_target_size()

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self._compute_target_size()

    def _index(self, page):
        return self._pages.index(page)

    def _update_sizes(self, ratio):
        """TODO"""
        if self.animation.currentTime() < self.animation.duration():
            new_size = round(self._targetSize * ratio)
            old_size = self._targetSize - new_size
            if new_size < self.layout().spacing():
                old_size -= self.layout().spacing()
            self._oldPage.set_height(max(0, old_size))
            self._newPage.set_height(new_size)
        else:
            super().setCurrentIndex(self._index(self._newPage))
            self._oldPage.finalize()
            self._newPage.finalize()

    def _compute_target_size(self):
        """TODO"""
        if not self.count():
            self._targetSize = 0
            return
        _, top, _, bottom = self.getContentsMargins()
        base_height = self._pages[0].button.sizeHint().height() + self.layout().spacing()
        self._targetSize = self.height() - top - bottom - base_height * self.count()

    def _button_clicked(self):
        """TODO"""
        button = self.sender()
        for i, page in enumerate(self._pages):
            if page.button == button:
                self.setCurrentIndex(i)
                return

    def _widget_destroyed(self):
        """TODO"""
        self._pages.remove(self.sender())
